package sdsi;

import sdsi.sexp.*;
import java.security.*;
/**
The class which implements acl's in SDSI/SPKI 2.0.  Acl's are currently considered
as SDSIObjects, but the grammar of SDSI/SPKI has changed so that acl's are
not specified.
@author Alex Morcos
@see SDSIObject
*/
public class Acl extends SDSIObject{
  SDSIPrincipal creatingPrincipal; //may not need to carry this around
  //may only need for creation
  AclEntry[] entries;
/**
  Creates a new Acl from a SexpList.  The SexpList is scanned and parsed and 
  the additional issuer information is also stored in the resulting Acl.
  
*/
  public Acl(SexpList l, SDSIPrincipal p) throws SexpParseException,SDSIException{
    super(l);
    if (p==null)
      throw new SDSIException("A creating principal must be defined to process an ACL");
    creatingPrincipal = p;
    if (!l.getType().equals("acl"))
      throw new SexpParseException("Not an ACL");
    int listSize = l.size();
    if (listSize >= 2) { //assumes zero acl entries is ok
      boolean versionDefined=false;
      if (srep.elementAt(1) instanceof SexpList && ((SexpList)srep.elementAt(1)).getType().equals("version"))
        versionDefined=true;
      int numEntries = versionDefined ? listSize-2 : listSize-1;
      entries = new AclEntry[numEntries];
      for (int i = 0;i < numEntries;i++) {
        if (srep.elementAt(i+listSize-numEntries) instanceof SexpList) {
          SexpList list = (SexpList)srep.elementAt(i+listSize-numEntries);
          entries[i] = new AclEntry(list);  
        }
        else
          throw new SexpParseException("ACL shouldn't contain a non-list");
      }
    }
  }
  public String getType() {
    return "acl";
  }
  public String toShortString() {
    return "ACL: " + entries.length + " entries";
  }
  public boolean contains(SDSIPublicKey applicant, Name applicantName, Tag request) {
    for (int i=0;i<entries.length;i++) {
      if (entries[i].allows(applicant, applicantName,request))
        return true;
    }
    return false;
  }
  public class AclEntry extends SDSIObject { //do I want to extend SDSIObject?
    //don't require that fields appear in order  , maybe i should
    Subject s;
    Validity v;
    Tag t;
    boolean delegate;
    public AclEntry(SexpList l) throws SexpParseException{
      super(l);
      boolean delegateDefined, commentDefined;
      delegateDefined = commentDefined = false;
      s = null;
      v = new Validity();
      t = null;
      delegate = false;
      if (!(l.getType().equals("entry")))
        throw new SexpParseException("Invalid entry in ACL");
      Sexp[] entryParts = l.arrayOfElements();
      for (int i=1;i < entryParts.length; i++) {
        Sexp element = entryParts[i];
        if (element instanceof SexpList) {
          SexpList list = (SexpList) element;
          String label = list.getType();
          if (label.equals("tag")) {
            if (t != null)
	            throw new SexpParseException("More than one tag in acl entry");
	          else
	            t = new Tag(list);
	        }
	        else if (label.equals("propagate")) {
	          if (delegateDefined)
	            throw new SexpParseException("More than one Delegation info in acl entry");
	          else {
	            delegateDefined = true;
	            if (list.size() != 1)
	              throw new SexpParseException("Delegation should be only a single word");
	            else
	              delegate = true;
	          }
	        }
	        else if (label.equals("not-after") || label.equals("not-before") || label.equals("online")) {
            v.update(list);
          }
          else if (label.equals("comment")) {
            if (commentDefined)
              throw new SexpParseException("More than one comment in acl entry");
            else {
              commentDefined = true;
              if (!(list.size() == 2) || !(list.elementAt(1) instanceof SexpString))
                throw new SexpParseException("Comment is of wrong form in acl entry");
            }
          }
          else {
            SDSIObject supposedSubject = SDSIObject.principalParse(list,creatingPrincipal);
            if (supposedSubject instanceof Subject)
              s = (Subject)supposedSubject;
            else
              throw new SexpParseException("Unknown sub-type in acl entry");
          }
        }
        else
      	  throw new SexpParseException("Shouldn't have a non-list in a cert");
      }
      if (s == null || t == null) 
        throw new SexpParseException("Need subject and tag in acl entry");
    }
    public String getType() {
      return "entry";
    }
    public String toShortString() {//may want to include subject tag info
      return "ACL entry";
    }
    public boolean allows(SDSIPublicKey applicant, Name applicantName, Tag request) {
      if (v.valid()) {
        if (t.intersect(request) == true) { 
          if (s instanceof SDSIPublicKey) {
            if (applicant.equals((SDSIPublicKey)s) && applicantName == null)
              return true;
          }
          if (s instanceof Name) {
            if (applicantName == null) 
              return false;
            Name subName = (Name) s;
            SDSIPrincipal subPrincipal = subName.getPrincipal();
            if (subPrincipal instanceof SDSIPublicKey) {
              if (((SDSIPublicKey)subPrincipal).equals(applicant)) {
                if (applicantName.sameNamesAs(subName))
                  return true;
              }
            }
            else { //subPrincipal instanceofHash
              try {
                Hash hashOfApp = new Hash(((Hash)subPrincipal).getHashID(),applicant);
                if (hashOfApp.equals((Hash)subPrincipal)) {
                  if (applicantName.sameNamesAs(subName))
                    return true;
                }
              }
              catch (NoSuchAlgorithmException e) {
                return false; //means can't process the acl
              }
            }
          }
        }
      }
      return false;
    }
  }
}
