package sdsi;

import sdsi.sexp.*;
import java.util.*;
/**The basic class which represents certificates in SDSI/SPKI 2.0.
  @author Alex Morcos
  @see Auth
  @see Def
  @see SDSIObject
  */
public abstract class Cert extends SDSIObject{
  SDSIPrincipal issuer;
  Subject s;
  Tag t;
  boolean delegate;
  Validity valid;
  boolean versionDefined,certDisplayDefined,commentDefined;
  /** Initializes the variables in the Cert, does
      not set up the SexpList representation.  This constructor
      is only called by subclasses which also do additional setup work*/
  protected Cert(SDSIPrincipal issuer, Subject s, Tag t, boolean delegate,
              Validity valid) 
  {
    this.issuer = issuer;
    this.s = s;
    this.t = t;
    this.delegate = delegate;
    this.valid = valid;
    versionDefined=certDisplayDefined=commentDefined=false;
  }
  /** Initializes the variables in the Cert, including setting up the
      SexpList representation.  Also checks that the list is really
      a cert.  This constructor is only called by subclasses which also 
      do additional setup work.
      @exception SexpParseException if the SexpList is not a correctly
      formatted SDSI certificate.*/
  protected Cert(SexpList l) throws SexpParseException{
    super(l);
    issuer = null;
    s = null;
    t = null;
    delegate = false;
    valid = new Validity();
    versionDefined = certDisplayDefined = commentDefined = false;
    if (!l.getType().equals("cert")) 
      throw new SexpParseException("Not a certificate");
    //need to correct to check for other fields and except on bad fields
  }
  /** Creates a new Cert. Scans the SexpList to
      determine whether the Cert is a Authorization or a Name Definition
      and then it returns the newly created appropriate sub-class.
      @param l the SexpList that contains the certificate
      @return a Cert object that represents the certificate
      @exception SexpParseException if the SexpList is not a correctly
      formatted SDSI certificate*/
  public static Cert create(SexpList l) throws SexpParseException{
    boolean authCert = false;
    Sexp[] certParts = l.arrayOfElements();
    SDSIObject issuerType = null;
    for (int j = 1; j < certParts.length; j++) {
      Sexp element = certParts[j];
      if (element instanceof SexpList) {
        SexpList list = (SexpList) element;
        String label = list.getType();
        if (label.equals("issuer")) {
          if (issuerType != null)
            throw new SexpParseException("More than one issuer in cert");
          if (list.size() != 2 || !(list.elementAt(1) instanceof SexpList))
            throw new SexpParseException("Issuer defined incorrectly");
          issuerType = SDSIObject.principalParse((SexpList)list.elementAt(1),null);
          if (issuerType instanceof Name) {
            authCert = false;
          }
          else if (issuerType instanceof SDSIPrincipal) {
            authCert = true;
          }
          else
            throw new SexpParseException("Issuer must be name or principal");
	      }
      }
    }
    if (issuerType == null) 
      throw new SexpParseException("Must have an issuer in cert");
    if (authCert)
      return new Auth(l,(SDSIPrincipal)issuerType);
    else
      return new Def(l,(Name)issuerType);
  }
  /**Parses a sub-list within the S-expression, this routine is called
    by the sub-class constructers.
    @exception SexpParseException if the sub-list does not correctly
    specify part of a Cert*/
  protected void parseItem(SexpList list) throws SexpParseException {
    String label = list.getType();
    if (label.equals("subject")) {
      if (s != null)
        throw new SexpParseException("More than one subject in cert");
      else if (issuer == null)
        throw new SexpParseException("Issuer must have been defined!!");
      else {
        if (!(list.size() == 2) || !(list.elementAt(1) instanceof SexpList))
          throw new SexpParseException("Subject of wrong form in cert");
        SDSIObject potentialSubject = SDSIObject.principalParse((SexpList)list.elementAt(1),issuer);
        if (potentialSubject instanceof Subject)
          s = (Subject)potentialSubject;
        else 
          throw new SexpParseException("Subject is not one of the allowed Subject objects");
      }
    }
    else if (label.equals("version")) {
      if (versionDefined)
        throw new SexpParseException("More than one version in cert");
      else {
        versionDefined = true;
        if (!(list.size() == 2) || !(list.elementAt(1) instanceof SexpString))
          throw new SexpParseException("Version is of wrong form in cert");
      }
    }
    else if (label.equals("display")) {
      if (certDisplayDefined)
        throw new SexpParseException("More than one display type for cert");
       else {
        certDisplayDefined = true;
         if (!(list.size() == 2) || !(list.elementAt(1) instanceof SexpString))
          throw new SexpParseException("Display type is of wrong form in cert");
      }
    }
    else if (label.equals("not-after") || label.equals("not-before") || label.equals("online")) {
      valid.update(list);
    }
    else if (label.equals("comment")) {
       if (commentDefined)
        throw new SexpParseException("More than comment in cert");
      else {
        commentDefined = true;
        if (!(list.size() == 2) || !(list.elementAt(1) instanceof SexpString))
          throw new SexpParseException("Comment is of wrong form in cert");
      }
    }
    else if (label.equals("issuer"))
      ;//already checked for this
    else
      throw new SexpParseException("Unknown sub-type in cert");
  }
  /**Tests if the Cert is currently valid.
  @return <code>true</code> if the cert is valid now, <code>false</code> otherwise.
  */
  public boolean valid() {
    return valid.valid();
  }
  /**Returns the issuer of this Cert.
  @return the SDSIPrincipal which is the issuer of this certificate.
  */
  public SDSIPrincipal getIssuer() {
    return issuer;
  }
  /**Returns the subject of this Cert.  All SDSIObjects that can be subjects
  of certificates implement the Subject interface.
  @return the SDSIObject which is the subject of this certificate.
  */
  public Subject getSubject() {
    return s;
  }
}  
