package sdsi;

import java.security.*;
import sdsi.sexp.*;
import java.math.*;
import java.io.*;
/**
 Class to represent SDSI objects.  The many subclasses are useful for providing specific 
 functionality inherent to individual types of objects.
 @author Alex Morcos
 @see Cert
 @see Acl
 @see Hash
 @see Name
 @see KeyHolder
 @see ObjectHash
 @see Principal
 @see Tag
 @see SDSIPrivateKey
 @see SDSISecretKey
 @see SDSIPublicKey
 @see Sequence
 @see SDSISignature
*/
public class SDSIObject implements Serializable{
 /* public interface Key {
    public String getAlgID();
  }
  */
  SexpList srep; //make private eventually!
  
  SDSIObject() {} 
  /**
  Create a new SDSIObject from a SexpList.
  @param s the SexpList that represents the SDSIObject.
  */
  public SDSIObject(SexpList s) {
    srep = s;
  }
   /** Generates the SDSI object that is represented by the SexpList in the argument.
       Parses the SexpList in the argument and returns the appropriate subclass
       of SDSIObject that has been initialized by the SexpList.  Is capable of parsing
       Acl's, KeyHolder's, and Name's that may require a creating princpal to be known.
       @param l the SexpLIst to be parsed.
       @param p the SDSIPrincipal that should be treated as having created the object
       @return the SDSI objecet represented
       @exception SexpParseException if the SexpList can not be succesfully parsed into a SDSIObject
   */
  public static SDSIObject principalParse(SexpList l, SDSIPrincipal p) throws SexpParseException{
    String type = l.getType();
    try {
      if (type.equals("acl"))
        return new Acl(l,p);
      if (type.equals("keyholder")) 
        return new KeyHolder(l,p);
      if (type.equals("name"))
        return new Name(l,p);
      else
        return SDSIObject.parse(l);
    }
    catch (SexpParseException e)
    {
      throw new SexpParseException("Error parsing " + type + ": " + e.getMessage(),l);
    }
     catch (SDSIException e)
    {
      throw new SexpParseException("SDSI error: "+e.getMessage(),l);//change
    }
  }
   /** Generates the SDSI object that is represented by the SexpList in the argument.
       Parses the SexpList in the argument and returns the appropriate subclass
       of SDSIObject that has been initialized by the SexpList.  
       @param l the SexpLIst to be parsed.
       @return the SDSI objecet represented
       @exception SexpParseException if the SexpList can not be succesfully parsed into a SDSIObject
   */
  private static SDSIObject parse(SexpList l) throws SexpParseException, SDSIException{
    String type = l.getType();
    if (type.equals("cert"))
      return Cert.create(l);
    if (type.equals("sequence"))
      return new Sequence(l);
    if (type.equals("public-key"))
      return SDSIPublicKey.create(l);
    if (type.equals("private-key"))
      return SDSIPrivateKey.create(l);
    if (type.equals("secret-key"))
      return new SDSISecretKey(l);
    if (type.equals("do"))
      return new Op(l);
    if (type.equals("signature"))
      return new SDSISignature(l);
    if (type.equals("hash"))
      return new Hash(l);
    /*if (type.equals("acl"))
      return new Acl(l);
    if (type.equals("keyholder")) //with this addition we might want to add a top level object interface
      return new KeyHolder(l);*/
    if (type.equals("object-hash"))
      return new ObjectHash(l);
    if (type.equals("tag"))
      return new Tag(l);
    /*if (type.equals("name"))
      return new Name(l);*/
    else
      return new SDSIObject(l); //eventually get rid of this
  }
  /**
     Returns a short descriptive string representing this SDSIObject.
     @return a short string description of this SDSIObject.
  */
  public String toShortString() {
    String stringRep = this.toString();
    return "Misc Object: " + stringRep.substring(0,(stringRep.length() < 10) ? stringRep.length() : 10);
  }
  /**
  Returns a string of this SDSIObject represented in canonical form.
  @return a string containing the canonical form of this SDSIObject.
  */
  public String toString() {
    return srep.toString();
  }
  /** 
  Returns a string representing this SDSIObject encoded as an S-Expression.
  The String contains only printable characters, is nicely formatted, and is a 
  valid S-Expression.
  @param width the maximum width of any line in the string.
  @return a nicely formatted string containing the S-Expression of this SDSIObject.
  */
  public String toReadableString(int width) {
    return srep.toReadableString(0, width, 0);
  }
  /**
  Returns the hash of this SDSIObject.  Hashes the canonical representation of this 
  SDSIObject and returns the resulting byte array.
  @param md the MessageDigest to use for the hash, it will be reset.
  @return the hash of the canonical representation of this SDSIObject as a byte array.
  */
  public byte[] getHash(MessageDigest md) {
    md.reset();
    md.update(srep.getCanonRep());
    return md.digest();
  }
  /**Updates a signature with the canonical representation of this SDSIObject.
  @param s the signature to update.
  @exception SignatureException if the argument is not a properly initialized Signature object.
  */
  public void sign(Signature s) throws SignatureException { //do I want to do this
    //or expose canonical rep idea somewhere, if so change hash also 
    s.update(srep.getCanonRep());
  }
  /**
  Returns a short string that represents the type of this SDSIObject.
  @return a string represent the type of SDSIObject we're dealing with.
  */
  public String getType() {
    return "SDSIObject";
  }
  /**Returns the S-Expression of this SDSIObject in canonical form as a SexpList
  @return the SexpList that encodes this SDSIObject in canonical form.
  */
  public SexpList getSrep() {
    return srep;
  }
}  
