package sdsi.sexp;

import java.io.*;

/** Abstract class to represent S-expressions.  There are two subclasses
    to represent byte-string or list S-expressions.
    @author Alex Morcos
    @see SexpString 
    @see SexpList
*/
public abstract class Sexp implements Serializable{

/**Tests if a character is whitespace.
   @param charToTest the character to be tested
   @return <code>true</code> if the character is whitespace, 
   <code>false</code> otherwise
*/
  public static final boolean isWhiteSpace(int charToTest) {
    if (charToTest == ' ' || charToTest == '\t' || charToTest == '\n' || charToTest == '\r')
      return true;
    else 
      return false;
  }
/**Tests if a character is a decimal digit.
   @param charToTest the character to be tested
   @return <code>true</code> if the character is a digit between 0 
   and 9 inclusive, <code>false</code> otherwise
*/  
  public static final boolean isDecimalDigit(int charToTest) {
    if (charToTest >= '0' && charToTest <= '9')
      return true;
    else
      return false;
  }
 /**Tests if a character is a hexadecimal digit.
   @param charToTest the character to be tested
   @return <code>true</code> if the character is a decimal digit or a letter 
   between 'a' and 'f' upper or lowercase, <code>false</code> otherwise
*/ 
  public static final boolean isHexDigit(int charToTest) {
    if  (isDecimalDigit(charToTest) || (charToTest >= 'a' && charToTest <= 'f') || (charToTest >= 'A' && charToTest <= 'F') )
      return true;
    else
      return false;
  }
 /**Tests if a character is a base 64 digit.
   @param charToTest the character to be tested
   @return <code>true</code> if the character is a decimal digit, a letter of the alphabet 
   or '+' or '/', <code>false</code> otherwise
*/
  public static final boolean isBase64Digit(int charToTest) {
    if (isDecimalDigit(charToTest) || (charToTest >= 'a' && charToTest <= 'z') || (charToTest >= 'A' && charToTest <= 'Z') || charToTest == '+' || charToTest == '/')
      return true;
    else
      return false;
  }
 /**Tests if a character is a token character.
   @param charToTest the character to be tested.
   @return <code>true</code> if the character is a legal token character 
   for an S-expression, <code>false</code> otherwise
*/ 
  public static final boolean isTokenChar(int charToTest) {
    if (isBase64Digit(charToTest) || charToTest=='.' || charToTest=='-' || charToTest=='=' || charToTest=='*' || charToTest==':' || charToTest=='_')
      return true;
    else
      return false;
  }
 /**
 Converts a base-64 digit to its corresponding integer value.
 @param b the byte representing the base-64 digit.
 @return the corresponing integer value between 0 and 63 as defined for 
 the S-expression base-64 encoding, or -1 if b is not valid base-64 digit.
 */
  public static final int base64ToInt(byte b) {
    if (b >= (byte) 'A' && b <= (byte) 'Z')
      return b - 'A';
    else if (b >= (byte) 'a' && b<= (byte) 'z')
      return b-'a'+26;
    else if (b >= (byte) '0' && b<= (byte) '9')
      return b-'0'+52;
    else if (b == '+') 
      return 62;
    else if (b == '/')
      return 63;
    else
      return -1;
  }
  /**
  Translates four base-64 digits into their corresponding three octet values.
  A base-64 string encodes three bytes as a sequence of four base-64 characters.
  This function takes four base-64 characters and recovers the three bytes being encoded.
  If the representation ends with one or two equal signs then the two or one, respectively,
  encoded bytes are returned.
  @param a,b,c,d the four base-64 digits to be decoded
  @return an array of bytes of length one, two, or three, containing the bytes that were encoded
  @exception SexpException if the bytes are not valid base-64 characters or do not correctly encode
  one, two, or three bytes.
  */
  public static final byte[] base64ToOctet(byte a,byte b,byte c, byte d) throws SexpException{
    byte[] answer;
    int a_num = 0;
    int b_num = 0;
    int c_num = 0;
    int d_num = 0;
    a_num = base64ToInt(a);
    b_num = base64ToInt(b);
    if (d == (byte) '=') {
      if (c == (byte) '=')  
        answer = new byte[1];
      else {
        c_num = base64ToInt(c);
        answer = new byte[2];
      }
    }
    else {
      c_num = base64ToInt(c);
      d_num = base64ToInt(d);
      answer = new byte[3];
    }
    if (a_num == -1 || b_num == -1 || c_num == -1 || d_num == -1)
      throw new SexpException("Invalid Base 64 character");
    answer[0] =(byte)((a_num<<2 & 0xFC) + (b_num>>>4 & 0x03));
    if (answer.length >= 2) {
      answer[1]=(byte)((b_num<<4 & 0xF0) + (c_num>>>2 & 0x0F));
      if (answer.length == 3)
        answer[2]=(byte)((c_num<<6 & 0xC0) + (d_num & 0x3F));
      else {
        if ((c_num & 0x03) != 0)
          throw new SexpException("Last Base 64 char must not contain extra bits");
      }
    }
    else {
      if ((b_num & 0x0F) != 0)
        throw new SexpException("Last Base 64 char must not contain extra bits");
    }
    return answer;
  }
  /**
  Provides a mapping from integers to the corresponding hexadecimal digit.
  */
  public static final char[] hexDigit = 
      {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
  /**
  Provides a mapping from integers to the corresponing base-64 digit.
  */
  public static final char[] base64Digit = {'A','B','C','D','E','F','G','H','I','J','K',
      'L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e',
      'f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y',
      'z','0','1','2','3','4','5','6','7','8','9','+','/'};
  /** Parses an S-expression out of the input stream.  The input stream is assumed to 
      contain an S-expression in canonical form.  Leaves the input stream at the end of 
      the S-expression that was parsed.
      @return the Sexp parsed.
      @param input an input stream conaining an S-expression.
      @exception SexpException if there was a problem in the parse.*/
  public static Sexp parse(PushbackInputStream input) throws SexpException { 
    int firstByte;
    try {
      while (Character.isWhitespace((char)(firstByte = input.read())))
        ;
      if (firstByte == -1) {
        return null;
      }
      else if (firstByte == '(') {
        return new SexpList(input);
      }
      else if (firstByte == '{') {
        byte[] newBytes = readBase64(input);
        return Sexp.parse(new PushbackInputStream(new ByteArrayInputStream(newBytes)));
      }
      else {
        input.unread(firstByte);
        return new SexpString(input);
      }
    }
    catch (IOException e) {
      System.err.println("There has been an input/output error");
      throw new SexpException("Failure in I/O");  //This should be changed
    }
  }
  /**
  Converts an array of bytes to a String.
  Assumes the byte array is encoded with the 8859_1 character set.
  @param b the byte array to be decoded
  @return the String whose bytes are in b
  */
  public static String toString(byte[] b) {
    String s = null;
    try {
      s = new String(b,"8859_1");
    }
    catch (UnsupportedEncodingException e){
      System.out.println("Unsupported Encoding, huh?");
    }
    return s;
  }
  /*Where do i use this?
  static boolean is_whitespace(int b) {
    if (b==' ' || b=='\t' || b=='\f' || b=='\r' || b=='\n') // Missing vertical tab
      return true;
    else
      return false;
    }
  /** Causes a human-understandble description of the Sexp to display
  public abstract void display();
  */
  /** Calculates the length of the canonical representation of this S-expression.
      @return the length of the canonical representation.*/
  public abstract int getCanonLen();
  /** Generates the canonical representation of this S-expression.
      @return a byte array containing exactly the canonical representation.*/
  public abstract byte[] getCanonRep();
 /** Returns a string representation of this S-expression.
     Possibly just the a String which represents the canonical-form.
     @return a string representation of this S-expression.
     */
  public String toString() {
    String s = null;
    try {
      s = new String(this.getCanonRep(),"8859_1");
    }
    catch (UnsupportedEncodingException e){
      System.out.println("Unsupported Encoding, huh?");
    }
    return s;
  }
  /** Returns a readable string representation of this S-expression.
      Given format parameters, returns a nicely formatted readable representation
      of this S-expression, using base-64 and hexadecimal representation to handle
      unpritnatble characters.
      @param offset the number of spaces that should appear before every line of the string.
      @param width the total width that the string should take up in characters.
      @param last the number of spaces that should be left at the end of every line (for possible closing parens).
      @return a nicely formatted string representation of this S-expression.
  */
  public abstract String toReadableString(int offset, int width, int last);
  abstract int getReadableLen();
 /** Returns a string representation of the canonical form of this S-expression.
     @return a string representation of the canoical form of this S-expression.
  */
  public String toVerbatimString() {
    String s = null;
    try {
      s = new String(this.getCanonRep(),"8859_1");
    }
    catch (UnsupportedEncodingException e){
      System.out.println("Unsupported Encoding, huh?");
    }
    return s;
  }
  private static byte[] readBase64(PushbackInputStream p) throws SexpException{  //this is almost the same code as
                                                        //SimpleString.readBase64Digits(p)
    try {
      int cur_byte=0;
      byte[] bytes = new byte[16];
      byte[] oldBytes = bytes;
      int bCount=0;
      int numEquals = 0;
      while (true) {
        cur_byte = p.read();
        if (cur_byte == -1)
          throw new SexpException("Error in Base 64 rep, not long enough");
        else if (cur_byte == '}') {
          int numDone;
          if ((numDone =(bCount % 4)) == 3) {
            numEquals++;
            bytes[bCount++] = (byte) '=';
          }
          else if (numDone ==2 ) {
            numEquals+=2;
            bytes[bCount++] = (byte) '=';
            bytes[bCount++] = (byte) '=';
          }
          else if (numDone == 0)
            ;
          else if (numDone == 1)
            throw new SexpException("Error in Base 64 rep, not long enough");
          break;
        }
        else if (cur_byte == '=') {
          int numDone;
          if ((numDone =(bCount % 4)) == 3 || numDone == 2) {
            numEquals++;
            bytes[bCount] = (byte) '=';
            bCount++;
            if (bCount == bytes.length) {
              bytes = new byte[bytes.length + 1];  
              System.arraycopy(oldBytes,0,bytes,0,oldBytes.length);
              oldBytes = bytes;
            }
          }
          else
            throw new SexpException("Error in Base 64 rep, too many equals");
        }
        else if (Sexp.isBase64Digit(cur_byte)) {
          if (numEquals > 0)
            throw new SexpException("Error in Base64 rep, = in the middle");
          bytes[bCount]=(byte) cur_byte;
          bCount++;
          if (bCount == bytes.length) {
            bytes = new byte[bytes.length *2];
            System.arraycopy(oldBytes,0,bytes,0,oldBytes.length);
            oldBytes = bytes;
          }
        }
        else if (Sexp.isWhiteSpace(cur_byte))
          continue;
        else
          throw new SexpException("Error, non-Base64 character in Base64 rep");
      }
      if (cur_byte != '}')
        throw new SexpException("Error, Base64 representation ended mistakenly"); // remove
      bytes = new byte[bCount];  //bCount should be mult of four here
      System.arraycopy(oldBytes,0,bytes,0,bCount);
      byte[] returnBytes = new byte[3*(bCount/4)-numEquals];
      byte[] threeBytes;
      for (int i=0; i< bCount/4;i++) {
        threeBytes = Sexp.base64ToOctet(bytes[4*i],bytes[4*i+1],bytes[4*i+2],bytes[4*i+3]);
        for (int j=0; j < threeBytes.length; j++)
          returnBytes[3*i+j] = threeBytes[j];
      }
      return returnBytes;
    }
    catch (IOException e) {
      throw new SexpException("There has been an input/output error reading a base64 rep in Sexp.parse");
    } 
  }
}
