package sdsi.sexp;

import java.util.*;
import java.io.*;
/**
   This class is used to represent S-Expressions which are not simple byte-strings.
   That is they are lists of other S-Expressions.
   @author Alex Morcos
   @see Sexp
*/
public class SexpList extends Sexp
  {
  Vector elements;
/**
   Create a new SexpList containing all the 
   Sexp's given in the array.
   @param expressions the array of expressions to be contained in the SexpList
*/
  public SexpList(Sexp[] expressions) {
    elements = new Vector(expressions.length);
    for (int i=0;i<expressions.length;i++)
      elements.addElement(expressions[i]);
  }
/**
   Create a new SexpList by parsing an S-Expression out of the input stream.
   Assumes that we have already parsed the first parenthesis of a list S-Expression.
   @param p the PushbackInputStream which contains the bytes of an S-Expression.
   @exception SexpException if there was a problem parsing the S-Expression
*/
  public SexpList(PushbackInputStream p) throws SexpException {
    elements = new Vector();
    Sexp next_exp;
    int cur_byte;
    try {
      while ((cur_byte = p.read()) != ')') { 
        if (Sexp.isWhiteSpace(cur_byte))
          continue;  //this should skip white space between strings
        else  
        {
          p.unread(cur_byte);
          next_exp = Sexp.parse(p); 
          if (next_exp == null)
            throw new SexpException("Does this ever happen: we parsed out a null sexp");         
          else {
            elements.addElement(next_exp);
          }
        }
      }
    }
    catch (IOException e) {
      throw new SexpException("Input/Output Error parsing S-Expression");
    }
  }
  /** Returns a string representing the first elemenent in the list.
      This string is equal to the ASCII translation of the content of the first
      Sexp in the list.  If the first Sexp is a list then
      the function returns "list"
      @return the first string in the list or "list"
  */
  public String getType() {
    Sexp first;
    if ((first=(Sexp)elements.firstElement()) instanceof SexpString) 
      return ((SexpString)first).stringContent();
    else
      return "list";  //indicates the first element of the list is a list
  }
  
  public int getCanonLen() {
    int len = 2;
    for (Enumeration e = elements.elements(); e.hasMoreElements();)
      len += ((Sexp)e.nextElement()).getCanonLen();
    return len;
  }
  
  public byte[] getCanonRep() {
    Sexp s;
    int s_len;
    int cur_pos = 1;
    byte[] canon = new byte[this.getCanonLen()];
    canon[0] = (byte) '(';
    for (Enumeration e = elements.elements(); e.hasMoreElements();) {
      s = (Sexp)e.nextElement();
      s_len = s.getCanonLen();
      System.arraycopy(s.getCanonRep(),0,canon,cur_pos,s_len);
      cur_pos += s_len;
    }
    canon[cur_pos] = (byte) ')';
    return canon;
  }
  int getReadableLen() {//assumes on one line
    int len = 1;
    Sexp[] elements = this.arrayOfElements();
    for (int i = 0;i<elements.length;i++) {
      len += elements[i].getReadableLen();
      len++;
    }
    return len;
  }
  public String toReadableString(int offset, int width, int last) {
    StringBuffer s = new StringBuffer();
    s.append("(");
    Sexp[] elements = this.arrayOfElements();
    if (elements.length == 0)
      return "()";
    boolean newLines = false;
    if (this.getReadableLen() > width-offset-last) {
      newLines = true;
      offset += 2;
      if (offset >= width)
        offset = 0;
    }
    for (int i=0;i<elements.length-1;i++) {
      if (newLines && i>0) {
        s.append("\n");
        for (int j=0;j<offset;j++)
          s.append(" ");
      }
      s.append(elements[i].toReadableString(offset,width,0) + " ");
    }
    if (newLines) {
      s.append("\n");
      for (int j=0;j<offset;j++)
        s.append(" ");
    }
    s.append(elements[elements.length-1].toReadableString(offset,width,last+1));
    s.append(")");
    return s.toString();
  }
/*
  //need to eliminate this function
  public void display() { //need to find a good way to do this!!
    byte[] disp = this.getCanonRep();
    System.out.write(disp,0,disp.length);
  } 
*/  
  /** Returns the Sexp located at the given index.  This is used when we know 
      that we want to get to the element at a particular location in the list
      @param index the index of the Sexp to return
      @return the Sexp located at index
  */
  public Sexp elementAt(int index) {
    return (Sexp)elements.elementAt(index);
  }
  /**  Returns the number of elements in this list
       @return the number of Sexp's in this SexpList.
  */  
  public int size() {
    return elements.size();
  }
  /**
     Returns an enumeration of the elements in the SexpList.  These will need to be
     cast to Sexp's.
     @return an enumberation of the elements in the SexpList.
  */
  public Enumeration elements(){//should return S-expressions! (I don't think it can)
    return elements.elements();
  }
  /** Puts all the Sexp's in the list into an array of the correct size and returns it
      @return an array of Sexp's from the list*/
  public Sexp[] arrayOfElements() {
    Sexp[] array = new Sexp[elements.size()];
    elements.copyInto(array);
    return array;
  }
  /** 
     Tests whether all the elements in this SexpList are simple strings or if there is a list.
     @return <code>false</code> if there is a SexpList inside this SexpList, <code>true</code> otherwise.
  */
  public boolean allElementsStrings() {
    for (Enumeration e = elements.elements();e.hasMoreElements();) {
      if (!(e.nextElement() instanceof SexpString))
        return false;
    }
    return true;
  }   
}
