package sdsi.sexp;

import java.io.*;
import java.util.*;
import java.security.*;  //so we can use MessageDigest.isEqual()
/**
   This class represents simple S-Expressions.  These are S-Expession which are just byte-strings.
   @author Alex Morcos
   @see Sexp
*/
public class SexpString extends Sexp{
  SimpleString display;
  SimpleString content;
/**
   Create a new SexpString with no display-hint and content equal to the String given.
   @param contents the contents of the SexpString.
*/
  public SexpString(String contents) {
    content = new SimpleString(contents);
    display = null;
  }
/**
   Create a new SexpString with the given display-hint and content.
   @param display_type the value of the display hint of the S-Expression.
   @param contents the content of the S-Expression.
*/
  public SexpString(String display_type, String contents) {
    display = new SimpleString(display_type);
    content = new SimpleString(contents);
  }
/**
   Create a new SexpString with no display-hint and content equal to the byte array given.
   @param contents the contents of the SexpString.
*/  
  public SexpString(byte[] contents) {
    content = new SimpleString(contents);
    display = null;
  }
/**
   Create a new SexpString with the given display-hint and content.
   @param display_type the value of the display hint of the S-Expression.
   @param contents the content of the S-Expression.
*/
  public SexpString(String display_type, byte[] contents) {
    display = new SimpleString(display_type);
    content = new SimpleString(contents);
  }
/**
   Create a new SexpString by parsing an S-Expression out of the input stream.
   Assumes that the S-Expression to parse is not a list.
   @param p the PushbackInputStream which contains the bytes of an S-Expression.
   @exception SexpException if there was a problem parsing the S-Expression
*/  
  public SexpString(PushbackInputStream p) throws SexpException {
    int cur_byte;
    try {
      if ((cur_byte = p.read()) == '[') {
        display = new SimpleString(p);
        if ((cur_byte = p.read()) != ']')
          throw new SexpException("Display type improperly encoded");
      }
      else  {
        p.unread(cur_byte);
        display = null;
      }
      content = new SimpleString(p);
    }
    catch (IOException e) {
      throw new SexpException("There has been an input/output error in SexpString");
    }
  }
/**
   Returns the SexpString as a String, without the display hint.
   @return a String representing the content of the SexpString.
*/
  public String stringContent() {
    return Sexp.toString(content.data);
  }
/**
   Returns the bytes of this SexpString, not including the display hint.
   @return a byte array containing the bytes of this SexpString.
*/
  public byte[] bytesContent() {
    return content.data;
  }
  /* don't think i need these two
  boolean canBeTokenString() {
    if (display == null)
      return this.content.canBeTokenString();
    else
      return (this.content.canBeTokenString() && this.display.canBeTokenString());
  }
  boolean canBeQuotedString() {
    if (display == null)
      return this.content.canBeQuotedString();
    else
      return (this.content.canBeQuotedString() && this.display.canBeQuotedString());
  }
  */
  public String toReadableString(int offset, int width, int last) {
    if (width <= offset + last + 1) {
      width = offset + last + 2;
    }
    if (display == null) {
      StringBuffer s = this.content.toReadableStringBuffer();
      if (s.length() > width - offset - last) {
        StringBuffer offsetPart = new StringBuffer("\n");
        for (int j=0;j<offset;j++)
          offsetPart.append(" ");
        for(int i=1;true;i++) {
          int next = (width + 1) * i - (offset+1);
          if (s.length() > next)
            s.insert(next,offsetPart.toString());
          else {
            if (s.length() > (next - last)) {
              s.append(offsetPart.toString().substring(0,offset));
              break;
            }
            else
              break;
          }
        }
      }
      return s.toString();
    }
    else {
      StringBuffer s = this.display.toReadableStringBuffer();
      s.insert(0,'[');
      s.append(']');
      StringBuffer offsetPart = new StringBuffer("\n");
      for (int j=0;j<offset;j++)
        offsetPart.append(" ");
      if (s.length() > width - offset) {
        for(int i=0;true;i++) {
          int next = width * i - offset;
          if (s.length() > next)
            s.insert(next,offsetPart.toString());
        }
      }
      s.append(offsetPart);
      StringBuffer c = this.content.toReadableStringBuffer();
      if (c.length() > width - offset) {
        for(int i=0;true;i++) {
          int next = width * i - offset;
          if (c.length() > next)
            c.insert(next,offsetPart.toString());
        }
      }
      s.append(c);
      return s.toString();
    }
  }
 int getReadableLen() {
    if (display != null)
      return display.getReadableLen() + 2 + content.getReadableLen();
    else
      return content.getReadableLen();
  }
  /*
  public void display() {
    if (display != null) {
      System.out.println("Display type is " + this.display.toString());
    }
    this.content.display();
  }
  */
  public int getCanonLen() {
    if (display != null)
      return display.getCanonLen() + 2 + content.getCanonLen();
    else
      return content.getCanonLen();
  }
  public byte[] getCanonRep() {
    if (display != null) {
      int disp_len = display.getCanonLen();
      byte[] canon = new byte[this.getCanonLen()];
      canon[0] = (byte) '[';
      System.arraycopy(display.getCanonRep(),0,canon,1,disp_len);
      canon[disp_len + 1] = (byte) ']';
      System.arraycopy(content.getCanonRep(),0,canon,disp_len+2,content.getCanonLen());
      return canon;
    }
    else 
      return content.getCanonRep();
  }
  /**
  Compares the string representation of two SexpStrings lexicographically.  Requires 
  that their display types are equal for them to be compared.  If the display types
  are equal then the contents of the SexpStrings are converted into verbatim string
  representations and compared without the leading lengths.
  @param other the SexpString to be compared.
  @return the value 0 if the argument SexpString is equal to this string; 
  a value less than 0 if this SexpString is lexicographically less
  than the SexpString argument; and a value greater than 0 if this 
  SexpString is lexicographically greater than the SexpString argument.
  @exception SexpException if the display types are not equal.
  */
  public int compareTo(SexpString other) throws SexpException{
    if (this.display == null) {
      if (other.display == null)
        return this.content.compareTo(other.content);
      else
        throw new SexpException("Display types must be equal");
    }
    else {
      if (this.display.equals(other.display))
        return this.content.compareTo(other.content);
      else
        throw new SexpException("Display types must be equal");
    }
  }
  /**
  Tests if the string representation of this SexpString starts with the string representation of
  another SexpString.  The test is performed on the verbatim string representations of SexpStrings
  with no display type and without the leading lengths.
  @param other the SexpString to serve as a prefix
  @return <code>true</code> if the character sequence represented by the argument is a prefix of 
  the character sequence represented by this SexpString; <code>false</code> otherwise.
  */
  public boolean startsWith(SexpString other) {
    if (this.display == null) {
      if (other.display == null)
        return this.content.startsWith(other.content);
      else
        return false;
    }
    if (this.display.equals(other.display))
      return this.content.startsWith(other.content);
    else
      return false;
  }
  /**
  Tests if two SexpStrings are equal.
  @param other the SexpString to test equality with
  @return <code>true</code> if both SexpStrings have the same display type and content,
  <code>false</code> otherwise.
  */
  public boolean equals(SexpString other) { //requires display types and contents are equal
    if (this.display == null) {
      if (other.display == null && this.content.equals(other.content))
        return true;
      else
        return false;
    }
    else 
      return (this.display.equals(other.display) && this.content.equals(other.content));
  }
}  
